(* May7.m

Date last modified:  24 June 2004

*)

(*
Contains definitions of various functions.
*)
(***************************************************************)
                   (* Error messges *)

TrCalc::AAAA="Optional TermList parameter must be a list or an integer."
TrCalc::BBBB="Term number is too large or too small."
TrCalc::CCCC="TermList parameter must not be an empty list."
TrCalc::DDDD="List element in TermList is invalid."
TrCalc::EEEE="Invalid permutation."
TrCalc::FFFF="Urgency parameter must be 1 or 2."
TrCalc::GGGG="Duplicate entries in TermList are not allowed."
TrCalc::HHHH="Permutation required."
TrCalc::IIII="Permutation of 1, 2, 3, required."
TrCalc::JJJJ="Specify 1, 2, and 3."
TrCalc::KKKK="Invalid parameter in RepeatFirstBianchiRule."
TrCalc::LLLL="Integer parameter in RepeatFirstBianchiRule must be positive."

(***************************************************************)

(* TracCalcValSelect:  Validate parameters for tractor calculus
     functions for which one has the option of selecting the terms
     on which the function is to act.

  Date last modified:  8 November 2003
    8 November 2003:  I checked this function over completely and
                   tested it completely.  I changed the numbers of
                   one or more of the error codes.
    16 July 2002:  I checked this function over carefully and ran
                   tests carefully.  See TestRule2.nb.

  Input:

    Operand:   a tensor expression
    TermList:  An integer or a list of integers (optional)

  Output:

    {error code, modified tensor expression} (i.e. a list)

  Some of our tractor calculus functions will have the ability to act
  on just selected terms of a tensor expression.  The present function
  will validate some of the parameters of such functions and return a
  list containing an error code and the desired terms.

  This function returns a list of length 2.  The first element of
  this list is an error code.  A value of zero indicates no error,
  and a nonzero value indicates an error.  The second element of the
  return list is either the sum of the terms requested by the caller
  (if no error), or Operand (if there is an error).

  TermList indicates which terms the caller has selected.  If the
  caller omits this parameter or codes "Omitted", then
  TracCalcValSelect returns Operand as the second element of the
  return list.

*)

Options[TracCalcValSelect] := { TermList->"Omitted" }

TracCalcValSelect[Operand_,opts___] :=

Module[{
  OldTensor,
  OpLength,
  ErrorFlag,
  LocalTermList,
  Holder,
  i
},

(

LocalTermList=(TermList/.{opts}/.Options[TracCalcValSelect]);

If[ (ToString[Head[LocalTermList]]=="String")&&
    (LocalTermList=="Omitted"),
      Return[{0,Operand}],
      Null
];

If[
  (ToString[Head[LocalTermList]] != "List")&&
  (ToString[Head[LocalTermList]] != "Integer"),
    (
    Message[TrCalc::AAAA];
    Return[{1,Operand}]
    ),
    Null
];      (* End If *)

If[ToString[Head[Operand]]=="Plus",
  OpLength=Length[Operand],
  OpLength=1
];

If[ToString[Head[LocalTermList]]=="Integer",
(**************************************************************
Special processing for the case in which LocalTermList is an Integer
**************************************************************)
  If[(LocalTermList>OpLength)||(LocalTermList<1),
    (
    Message[TrCalc::BBBB];
    Return[{2,Operand}]
    ),
    If[OpLength==1,
      Return[{0,Operand}],
      Return[{0,Operand[[LocalTermList]]}]
    ]
  ],
(**********************************************************
Special processing for the case in which LocalTermList is a List
**********************************************************)
  (
  If[Length[LocalTermList]==0,
    (
    Message[TrCalc::CCCC];
    Return[{3,Operand}]
    ),
    Null
  ];
  ErrorFlag=False;
  Do[
    If[ (ToString[Head[LocalTermList[[i]]]] != "Integer")||
        (LocalTermList[[i]]>OpLength)||
        (LocalTermList[[i]]<1),
    ErrorFlag=True,
    Null
    ],
    {i,1,Length[LocalTermList]}
  ]; (* End Do *)
  If[ErrorFlag,
    (
    Message[TrCalc::DDDD];
    Return[{4,Operand}]
    ),
    Null
  ];  (* End If ErrorFlag *)
  If[(Length[LocalTermList]>1),
    (
    ErrorFlag=False;
    Holder=Sort[LocalTermList];
    Do[
      If[Holder[[i]]==Holder[[i+1]],
        ErrorFlag=True,
        Null
      ],
    {i,1,Length[LocalTermList]-1}
    ];
    If[ErrorFlag,
      (
      Message[TrCalc::GGGG];
      Return[{7,Operand}]
      ),
      Null
    ];  (* End If ErrorFlag *)
    ),
    Null
  ];  (* End If Length > 1 *)
  If[OpLength==1,
    Return[{0,Operand}],
    Null
  ];
  OldTensor=0;
  Do[
    OldTensor=(OldTensor+Operand[[LocalTermList[[i]]]]),
    {i,1,Length[LocalTermList]}
  ];
  Return[{0,OldTensor}]
  )  (* End List case *)
]; (* End If *)

)  (* End body of Module *)
]; (* End of Module command *)

(***************************************************************)
(*                         SwapRiem                            *)
(***************************************************************)
(* SwapRiem:  Swap (interchange) copies of Riemannian curvature tensor.

Date last modified:  16 July 2002
    16 July 2002:  I checked this function over carefully and ran
                   tests carefully.  See TestRule2.nb.

Input:

Operand:  A tensor expression (which may include tractors)

TermList:  Parameter is optional.  TermList is a positive integer or a
  list of such integers indicating the terms to which the user wishes to
  apply the function.  If omitted, or if the user codes "Omitted", then
  SwapRiem acts on all of Operand.

Urgency:  Parameter is optional.  Default value is 2, which means that
  SwapRiem will apply the SuperSimplify command to its intermediate
  result.  The other option available to the caller is to specify
  Urgency=1.  In this case SwapRiem will not call SuperSimplify.  This
  speeds up the computation but reduces the likelihood that SwapRiem
  will successfully interchange the copies of the Riemannian curvature
  tensor.  See remarks below.

Output:

A tensor expression equivalent to Operand.

At times one may want to change the order in which Ricci lists the
factors of some term of a tensor expression.  This problem occurs when
the same tensor appears more than once as a factor of a given term.
This function attempts to interchange two copies of the
undifferentiated Riemannian curvature tensor which appear as factors.
If a term has more than two such factors, then the outcome of this
function is unpredictable.  In any case, this function will return a
tensor-tractor expression which is equivalent to Operand.  The user,
of course, must specify legitimate values for the parameters.

Ricci has certain procedures for deciding upon the order in which it
arranges factors of tensor expressions.  These rules may prevent this
function from successfully arranging the factors in the order
requested by the caller.  In any case, this function always returns a
tensor expression which is equivalent to Operand.

*)

Options[SwapRiem] := {TermList->"Omitted",Urgency->2};

SwapRiem[Operand_,opts___] :=

Module[{
  ValResult,
  Temp,
  OldTensor,
  NewTensor,
  ResultTensor,
  LocalUrgency,
  LocalTermList
},
(
  LocalUrgency=(Urgency/.{opts}/.Options[SwapRiem]);
  LocalTermList=(TermList/.{opts}/.Options[SwapRiem]);
  
  If[(ToString[Head[LocalUrgency]] != "Integer")||
     (LocalUrgency<1)||
     (LocalUrgency>2),
    (
    Message[TrCalc::FFFF];
    Return[Operand]
    ),
    Null
  ];

  ValResult=TracCalcValSelect[Operand,TermList->LocalTermList];

  If[ValResult[[1]]!=0,
    Return[Operand],
    (
    OldTensor=ValResult[[2]];
    Temp=OldTensor;
    Temp=(Temp/.RmRmToRmAA);
    Temp=TensorSimplify[Temp];
    If[LocalUrgency==2,
      Temp=SuperSimplify[Temp],
      Null];
    Temp=(Temp/.AAToRm);
    Temp=NewDummy[Temp];
    NewTensor=Temp;
    ResultTensor=(Operand+NewTensor-OldTensor//TensorExpand);
    Return[ResultTensor]
    )
  ]   (* End If *)
)     (* End body of Module *)
];    (* End Module *)

(************************************************************)
(*                      PermuteRiem                         *)
(************************************************************)
(*
Date last modified:  16 July 2002
    16 July 2002:  I checked this function over carefully and ran
                   tests carefully.  See TestRule2.nb.

Input:

Operand:  a tensor expression
FactorPermutation:  a permutation of 1, 2, 3
TermList:  Parameter is optional.  It is a positive integer or a list
  of such integers indicating the term or terms on which the function
  is to act.
Urgency:  Parameter is optional.  Default value is 2, which causes
  PermuteRiem to apply the SuperSimplify command to its intermediate
  result.  A value of 1 for this parameter speeds up the computation
  but also increases the likelihood of the function not producing the
  desired rearrangement of factors.

Ouput:

ResultTensor (always equivalent to operand)

Function receives a tensor expression in which at least one term has
three undifferentiated Riemannian curvature tensor factors.  Function
acts on the terms indicated by TermList.  If TermList is omitted, then
function acts on the entire Operand.  Function attempts to rearrange
the order of the Riemannian curvature tensor factors according to the
given pertutation.  The permutation must be a permutation of the
numbers 1, 2, and 3.  The attempted reordering is done as in the
following example:

{2,3,1} means move factor 2 to slot 1, factor 3 to slot 2, and factor
1 to slot 3.

If three undifferentiated copies of the Riamannian curvature tensor
are not present or are not in the location or locations specified by
TermList, or if one or more copies of the Riemannian curvature tensor
are differentiated, this function will do no harm.  It will always
return a tensor expression equivalent to Operand.

Ricci has certain procedures for deciding upon the order in which it
arranges factors of tensor expressions.  These rules may prevent this
function from successfully arranging the factors in the order
requested by the caller.  In any case, this function always returns a
tensor expression which is equivalent to Operand.

*)

DefineTensor[AAAARm,4,Symmetries->RiemannSymmetries];
DefineTensor[BBBBRm,4,Symmetries->RiemannSymmetries];
DefineTensor[CCCCRm,4,Symmetries->RiemannSymmetries];

DefineRule[ReturnRiemann,AAAARm,Rm,NewRule->True];
DefineRule[ReturnRiemann,BBBBRm,Rm];
DefineRule[ReturnRiemann,CCCCRm,Rm];

Options[PermuteRiem] = {Urgency->2,TermList->"Omitted"};

PermuteRiem[Operand_,FactorPermutation_,
  opts___] :=

Module[
{
  ValResult,
  LocalRule,
  LocalTermList,
  LocalUrgency,
  Temp,
  OldTensor,
  NewTensor,
  ResultTensor
},
(
  LocalUrgency=(Urgency/.{opts}/.Options[PermuteRiem]);
  LocalTermList=(TermList/.{opts}/.Options[PermuteRiem]);

  If[(ToString[Head[LocalUrgency]] != "Integer")||
     (LocalUrgency<1)||
     (LocalUrgency>2),
    (
    Message[TrCalc::FFFF];
    Return[Operand]
    ),
    Null
  ];

  ValResult=TracCalcValSelect[Operand,TermList->LocalTermList];

  If[ValResult[[1]] != 0, Return[Operand], Null];

  If[(ToString[Head[FactorPermutation]] != "List")||
     (Length[FactorPermutation] != 3)||
     (ToString[Head[FactorPermutation[[1]]]] != "Integer")||
     (ToString[Head[FactorPermutation[[2]]]] != "Integer")||
     (ToString[Head[FactorPermutation[[3]]]] != "Integer")||
     (Sort[FactorPermutation] != {1,2,3}),
    (
    Message[TrCalc::EEEE];
    Return[Operand]
    ),
    Null
  ];

  Which[
    FactorPermutation=={3,2,1},
      DefineRule[LocalRule,
                 Rm[L[i1],L[i2],L[i3],L[i4]]*
                 Rm[L[j1],L[j2],L[j3],L[j4]]*
                 Rm[L[k1],L[k2],L[k3],L[k4]],
             CCCCRm[L[i1],L[i2],L[i3],L[i4]]*
             BBBBRm[L[j1],L[j2],L[j3],L[j4]]*
             AAAARm[L[k1],L[k2],L[k3],L[k4]],
                 NewRule->True,Quiet->True],
    FactorPermutation=={3,1,2},
      DefineRule[LocalRule,
                 Rm[L[i1],L[i2],L[i3],L[i4]]*
                 Rm[L[j1],L[j2],L[j3],L[j4]]*
                 Rm[L[k1],L[k2],L[k3],L[k4]],
             BBBBRm[L[i1],L[i2],L[i3],L[i4]]*
             CCCCRm[L[j1],L[j2],L[j3],L[j4]]*
             AAAARm[L[k1],L[k2],L[k3],L[k4]],
                 NewRule->True,Quiet->True],
    FactorPermutation=={2,1,3},
      DefineRule[LocalRule,
                 Rm[L[i1],L[i2],L[i3],L[i4]]*
                 Rm[L[j1],L[j2],L[j3],L[j4]]*
                 Rm[L[k1],L[k2],L[k3],L[k4]],
             BBBBRm[L[i1],L[i2],L[i3],L[i4]]*
             AAAARm[L[j1],L[j2],L[j3],L[j4]]*
             CCCCRm[L[k1],L[k2],L[k3],L[k4]],
                 NewRule->True,Quiet->True],
    FactorPermutation=={2,3,1},
      DefineRule[LocalRule,
                 Rm[L[i1],L[i2],L[i3],L[i4]]*
                 Rm[L[j1],L[j2],L[j3],L[j4]]*
                 Rm[L[k1],L[k2],L[k3],L[k4]],
             CCCCRm[L[i1],L[i2],L[i3],L[i4]]*
             AAAARm[L[j1],L[j2],L[j3],L[j4]]*
             BBBBRm[L[k1],L[k2],L[k3],L[k4]],
                 NewRule->True,Quiet->True],
    FactorPermutation=={1,3,2},
      DefineRule[LocalRule,
                 Rm[L[i1],L[i2],L[i3],L[i4]]*
                 Rm[L[j1],L[j2],L[j3],L[j4]]*
                 Rm[L[k1],L[k2],L[k3],L[k4]],
             AAAARm[L[i1],L[i2],L[i3],L[i4]]*
             CCCCRm[L[j1],L[j2],L[j3],L[j4]]*
             BBBBRm[L[k1],L[k2],L[k3],L[k4]],
                 NewRule->True,Quiet->True],
    True,
      Return[Operand]
  ];    (* End Which *)

  OldTensor=ValResult[[2]];
  Temp=OldTensor;
  Temp=(Temp/.LocalRule);
  Temp=TensorSimplify[Temp];
  If[LocalUrgency==2,
    Temp=SuperSimplify[Temp],
    Null
  ];
  Temp=(Temp/.ReturnRiemann);
  Temp=NewDummy[Temp];
  NewTensor=Temp;
  ResultTensor=(Operand+NewTensor-OldTensor//TensorExpand);
  Return[ResultTensor]

)  (* End body of Module *)
]; (* End Module *)

(************************************************************)
(*                     PermuteNabRc                         *)
(************************************************************)
(*
Date last modified:  22 July 2002

On 22 July 2002, I inspected and tested this function carefully.  See
TestRule3.nb.

Input:

Operand:  a tensor expression
FactorPermutation:  a permutation of 1, 2, 3
TermList:  Parameter is optional.  It is a positive integer or a list
  of such integers indicating the term or terms on which the function
  is to act.
Urgency:  Parameter is optional.  Default value is 2, which causes
  PermuteNabRc to apply the SuperSimplify command to its intermediate
  result.  A value of 1 for this parameter speeds up the computation
  but also increases the likelihood of the function not producing the
  desired rearrangement of factors.

Ouput:

ResultTensor (always equivalent to operand)

Function receives a tensor expression in which at least one term has
three differentiated Ricci tensor factors.  Each of these three
factors is assumed to have only one derivative applied to it.
Function acts on the terms indicated by TermList.  If TermList is
omitted, then function acts on the entire Operand.  Function attempts
to rearrange the order of the differentiated Ricci curvature tensor
factors according to the given pertutation.  The permutation must be a
permutation of the numbers 1, 2, and 3.  The attempted reordering is
done as in the following example:

{2,3,1} means move factor 2 to slot 1, factor 3 to slot 2, and factor
1 to slot 3.

If the three differentiated copies of the Ricci tensor are not present
or are not in the location or locations specified by TermList, or if
one or more copies of the Ricci tensor are differentiated more than
once, this function will do no harm.  It will always return a tensor
expression equivalent to Operand.

Ricci has certain procedures for deciding upon the order in which it
arranges factors of tensor expressions.  These rules may prevent this
function from successfully arranging the factors in the order
requested by the caller.  In any case, this function always returns a
tensor expression which is equivalent to Operand.

*)

DefineTensorSymmetries[PNabRcSymmetries,{{2,1,3},+1}];

DefineTensor[AAAARicciDD,3,Symmetries->PNabRcSymmetries];
DefineTensor[BBBBRicciDD,3,Symmetries->PNabRcSymmetries];
DefineTensor[CCCCRicciDD,3,Symmetries->PNabRcSymmetries];

DefineRule[ReturnNabRc,AAAARicciDD[L[i],L[j],L[k]],Rc[L[i],L[j]][L[k]],
  NewRule->True];
DefineRule[ReturnNabRc,BBBBRicciDD[L[i],L[j],L[k]],Rc[L[i],L[j]][L[k]]];
DefineRule[ReturnNabRc,CCCCRicciDD[L[i],L[j],L[k]],Rc[L[i],L[j]][L[k]]];

Options[PermuteNabRc] = {Urgency->2,TermList->"Omitted"};

PermuteNabRc[Operand_,FactorPermutation_,opts___] :=

Module[
{
  ValResult,
  LocalRule,
  LocalTermList,
  LocalUrgency,
  Temp,
  OldTensor,
  NewTensor,
  ResultTensor
},
(
  LocalUrgency=(Urgency/.{opts}/.Options[PermuteNabRc]);
  LocalTermList=(TermList/.{opts}/.Options[PermuteNabRc]);

  If[(ToString[Head[LocalUrgency]] != "Integer")||
     (LocalUrgency<1)||
     (LocalUrgency>2),
    (
    Message[TrCalc::FFFF];
    Return[Operand]
    ),
    Null
  ];

  ValResult=TracCalcValSelect[Operand,TermList->LocalTermList];

  If[ValResult[[1]] != 0, Return[Operand], Null];

  If[(ToString[Head[FactorPermutation]] != "List")||
     (Length[FactorPermutation] != 3)||
     (ToString[Head[FactorPermutation[[1]]]] != "Integer")||
     (ToString[Head[FactorPermutation[[2]]]] != "Integer")||
     (ToString[Head[FactorPermutation[[3]]]] != "Integer")||
     (Sort[FactorPermutation] != {1,2,3}),
    (
    Message[TrCalc::EEEE];
    Return[Operand]
    ),
    Null
  ];

  Which[
    FactorPermutation=={3,2,1},
      DefineRule[LocalRule,
                 Rc[L[i1],L[i2]][L[i3]]*
                 Rc[L[j1],L[j2]][L[j3]]*
                 Rc[L[k1],L[k2]][L[k3]],
        CCCCRicciDD[L[i1],L[i2],L[i3]]*
        BBBBRicciDD[L[j1],L[j2],L[j3]]*
        AAAARicciDD[L[k1],L[k2],L[k3]],
                 NewRule->True,Quiet->True],
    FactorPermutation=={3,1,2},
      DefineRule[LocalRule,
                 Rc[L[i1],L[i2]][L[i3]]*
                 Rc[L[j1],L[j2]][L[j3]]*
                 Rc[L[k1],L[k2]][L[k3]],
        BBBBRicciDD[L[i1],L[i2],L[i3]]*
        CCCCRicciDD[L[j1],L[j2],L[j3]]*
        AAAARicciDD[L[k1],L[k2],L[k3]],
                 NewRule->True,Quiet->True],
    FactorPermutation=={2,1,3},
      DefineRule[LocalRule,
                 Rc[L[i1],L[i2]][L[i3]]*
                 Rc[L[j1],L[j2]][L[j3]]*
                 Rc[L[k1],L[k2]][L[k3]],
        BBBBRicciDD[L[i1],L[i2],L[i3]]*
        AAAARicciDD[L[j1],L[j2],L[j3]]*
        CCCCRicciDD[L[k1],L[k2],L[k3]],
                 NewRule->True,Quiet->True],
    FactorPermutation=={2,3,1},
      DefineRule[LocalRule,
                 Rc[L[i1],L[i2]][L[i3]]*
                 Rc[L[j1],L[j2]][L[j3]]*
                 Rc[L[k1],L[k2]][L[k3]],
        CCCCRicciDD[L[i1],L[i2],L[i3]]*
        AAAARicciDD[L[j1],L[j2],L[j3]]*
        BBBBRicciDD[L[k1],L[k2],L[k3]],
                 NewRule->True,Quiet->True],
    FactorPermutation=={1,3,2},
      DefineRule[LocalRule,
                 Rc[L[i1],L[i2]][L[i3]]*
                 Rc[L[j1],L[j2]][L[j3]]*
                 Rc[L[k1],L[k2]][L[k3]],
        AAAARicciDD[L[i1],L[i2],L[i3]]*
        CCCCRicciDD[L[j1],L[j2],L[j3]]*
        BBBBRicciDD[L[k1],L[k2],L[k3]],
                 NewRule->True,Quiet->True],
    True,
      Return[Operand]
  ];    (* End Which *)

  OldTensor=ValResult[[2]];
  Temp=OldTensor;
  Temp=(Temp/.LocalRule);
  Temp=TensorSimplify[Temp];
  If[LocalUrgency==2,
    Temp=SuperSimplify[Temp],
    Null
  ];
  Temp=(Temp/.ReturnNabRc);
  Temp=NewDummy[Temp];
  NewTensor=Temp;
  ResultTensor=(Operand+NewTensor-OldTensor//TensorExpand);
  Return[ResultTensor]

)  (* End body of Module *)
]; (* End Module *)

(***************************************************************)
(*                        SwapNabRc                            *)
(***************************************************************)
(* SwapNabRc:  Swap (interchange) copies of the first-order covariant
               derivative of the Ricci tensor.

Date last modified:  17 July 2002

On 17 July 2002, I checked this function over carefully.  I ran tests
in TestRule2.nb.

Input:

Operand:  A tensor expression (which may include tractors)

TermList:  Parameter is optional.  TermList is a positive integer or a
  list of such integers indicating the terms to which the user wishes to
  apply the function.  If omitted, or if the user codes "Omitted", then
  SwapRiem acts on all of Operand.

Urgency:  Parameter is optional.  Default value is 2, which means that
  SwapRiem will apply the SuperSimplify command to its intermediate
  result.  The other option available to the caller is to specify
  Urgency=1.  In this case SwapRiem will not call SuperSimplify.  This
  speeds up the computation but reduces the likelihood that SwapRiem
  will successfully interchange the copies of the Riemannian curvature
  tensor.  See remarks below.

Output:

A tensor expression equivalent to Operand.

At times one may want to change the order in which Ricci lists the
factors of some term of a tensor expression.  This problem occurs when
the same tensor appears more than once as a factor of a given term.
This function attempts to interchange two copies of the first-order
derivative of the Ricci curvature tensor which appear as factors.  If
a term has more than two such factors, then the outcome of this
function is unpredictable.  In any case, this function will return a
tensor-tractor expression which is equivalent to Operand.  The user,
of course, must specify legitimate values for the parameters.

Ricci has certain procedures for deciding upon the order in which it
arranges factors of tensor expressions.  These rules may prevent this
function from successfully arranging the factors in the order
requested by the caller.  In any case, this function always returns a
tensor expression which is equivalent to Operand.

*)

Options[SwapNabRc] := {TermList->"Omitted",Urgency->2};

SwapNabRc[Operand_,opts___] :=

Module[{
  ValResult,
  Temp,
  OldTensor,
  NewTensor,
  ResultTensor,
  LocalUrgency,
  LocalTermList
},
(
  LocalUrgency=(Urgency/.{opts}/.Options[SwapNabRc]);
  LocalTermList=(TermList/.{opts}/.Options[SwapNabRc]);
  
  If[(ToString[Head[LocalUrgency]] != "Integer")||
     (LocalUrgency<1)||
     (LocalUrgency>2),
    (
    Message[TrCalc::FFFF];
    Return[Operand]
    ),
    Null
  ];

  ValResult=TracCalcValSelect[Operand,TermList->LocalTermList];

  If[ValResult[[1]]!=0,
    Return[Operand],
    (
    OldTensor=ValResult[[2]];
    Temp=OldTensor;
    Temp=(Temp/.RcDRcDToRcDAA);
    Temp=TensorSimplify[Temp];
    If[LocalUrgency==2,
      Temp=SuperSimplify[Temp],
      Null];
    Temp=(Temp/.AAToRcD);
    Temp=NewDummy[Temp];
    NewTensor=Temp;
    ResultTensor=(Operand+NewTensor-OldTensor//TensorExpand);
    Return[ResultTensor]
    )
  ]   (* End If *)
)     (* End body of Module *)
];    (* End Module *)

(***************************************************************)
(*                        SwapRicci                            *)
(***************************************************************)
(* SwapRicci:  Swap (interchange) copies of the undifferentiated Ricci
               tensor.

Date last modified:  19 July 2002

On 22 July 2002, I checked this function over carefully.  I carefully
ran tests.  See TestRule3.nb.

Input:

Operand:  A tensor expression (which may include tractors)

TermList:  Parameter is optional.  TermList is a positive integer or a
  list of such integers indicating the terms to which the user wishes to
  apply the function.  If omitted, or if the user codes "Omitted", then
  SwapRicci acts on all of Operand.

Urgency:  Parameter is optional.  Default value is 2, which means that
  SwapRicci will apply the SuperSimplify command to its intermediate
  result.  The other option available to the caller is to specify
  Urgency=1.  In this case SwapRicci will not call SuperSimplify.  This
  speeds up the computation but reduces the likelihood that SwapRicci
  will successfully interchange the copies of the Riemannian curvature
  tensor.  See remarks below.

Output:

A tensor expression equivalent to Operand.

At times one may want to change the order in which Ricci lists the
factors of some term of a tensor expression.  This problem occurs when
the same tensor appears more than once as a factor of a given term.
This function attempts to interchange two copies of the
undifferentiated Ricci curvature tensor which appear as factors.  If a
term has more than two such factors, then the outcome of this function
is unpredictable.  In any case, this function will return a
tensor-tractor expression which is equivalent to Operand.  The user,
of course, must specify legitimate values for the parameters.

Ricci has certain procedures for deciding upon the order in which it
arranges factors of tensor expressions.  These rules may prevent this
function from successfully arranging the factors in the order
requested by the caller.  In any case, this function always returns a
tensor expression which is equivalent to Operand.

*)

Options[SwapRicci] := {TermList->"Omitted",Urgency->2};

SwapRicci[Operand_,opts___] :=

Module[{
  ValResult,
  Temp,
  OldTensor,
  NewTensor,
  ResultTensor,
  LocalUrgency,
  LocalTermList
},
(
  LocalUrgency=(Urgency/.{opts}/.Options[SwapRicci]);
  LocalTermList=(TermList/.{opts}/.Options[SwapRicci]);
  
  If[(ToString[Head[LocalUrgency]] != "Integer")||
     (LocalUrgency<1)||
     (LocalUrgency>2),
    (
    Message[TrCalc::FFFF];
    Return[Operand]
    ),
    Null
  ];

  ValResult=TracCalcValSelect[Operand,TermList->LocalTermList];

  If[ValResult[[1]]!=0,
    Return[Operand],
    (
    OldTensor=ValResult[[2]];
    Temp=OldTensor;
    Temp=(Temp/.RcRcToFake);
    Temp=TensorSimplify[Temp];
    If[LocalUrgency==2,
      Temp=SuperSimplify[Temp],
      Null];
    Temp=(Temp/.AAToPlainRc);
    Temp=NewDummy[Temp];
    NewTensor=Temp;
    ResultTensor=(Operand+NewTensor-OldTensor//TensorExpand);
    Return[ResultTensor]
    )
  ]   (* End If *)
)     (* End body of Module *)
];    (* End Module *)


(***********************************************************)
(*                        PermWeyl                         *)
(***********************************************************)
(* Function to permute copies of the Weyl tensor in a given
   term.  This function assumes that there are three copies
   of the Weyl tensor in the term or terms of interest.  If this
   is not the case, the function will not do any damage.

Originated on 2 November 2003
Date last modified:  8 November 2003
      I checked and tested this function carefully on 8 November 2003.

Required parameters:

(1) A tensor expression
(2) A permutation of the numbers 1, 2, and 3 (in the form of a
    list)

Optional parameters:

(3) A list or a single integer indicating which term or terms of the
    tensor expression the function is to act on.  Default value is
    "Omitted", which indicates that the function is to act on all
    terms of the expression.  The caller should code this as
    TermList->{2,5,9} (for example) or TermList->3 (for example).
(4) RetWeyl flag.  If set to "True", then, upon completion of
    the main work, the function will replace WWAA, WWBB, and WWCC with
    the Weyl tensor.  Default value is "True".

Return value:

A tensor expression which is equivalent to parameter (1).

*)

DefineTensor[WWAA,4,Symmetries->RiemannSymmetries];
DefineTensor[WWBB,4,Symmetries->RiemannSymmetries];
DefineTensor[WWCC,4,Symmetries->RiemannSymmetries];
DefineRule[ReturnThreeWeyl,WWAA,Weyl,NewRule->True];
DefineRule[ReturnThreeWeyl,WWBB,Weyl];
DefineRule[ReturnThreeWeyl,WWCC,Weyl];

(**** Main function definition ****)

Options[PermWeyl] := {RetWeyl->True,TermList->"Omitted"};

PermWeyl[TenExp_,ThreePerm_,opts___]:=

Module[
{
FormFlag,
LocalTermList,
NewTensor,
OldTensor,
PermRule,
ReturnValue,
Temp,
ValResult
},

(

ReturnValue=TenExp;

FormFlag=(RetWeyl/.{opts}/.Options[PermWeyl]);
LocalTermList=(TermList/.{opts}/.Options[PermWeyl]);

ValResult=TracCalcValSelect[TenExp,TermList->LocalTermList];

If[ValResult[[1]]!=0,Return[ReturnValue],Null];

OldTensor=ValResult[[2]];

If[ToString[Head[ThreePerm]]!="List",
(Message[TrCalc::HHHH];Return[ReturnValue]),
Null];

If[Length[ThreePerm]!=3,
(Message[TrCalc::IIII];Return[ReturnValue]),
Null];

If[Sort[ThreePerm]!={1,2,3},
(Message[TrCalc::JJJJ];Return[ReturnValue]),
Null];

Which[

ThreePerm=={1,2,3},
DefineRule[PermRule,Weyl[L[i1],L[i2],L[i3],L[i4]]*
                    Weyl[L[j1],L[j2],L[j3],L[j4]]*
                    Weyl[L[k1],L[k2],L[k3],L[k4]],
                    WWAA[L[i1],L[i2],L[i3],L[i4]]*
                    WWBB[L[j1],L[j2],L[j3],L[j4]]*
                    WWCC[L[k1],L[k2],L[k3],L[k4]],
  NewRule->True,Quiet->True],

ThreePerm=={1,3,2},
DefineRule[PermRule,Weyl[L[i1],L[i2],L[i3],L[i4]]*
                    Weyl[L[j1],L[j2],L[j3],L[j4]]*
                    Weyl[L[k1],L[k2],L[k3],L[k4]],
                    WWAA[L[i1],L[i2],L[i3],L[i4]]*
                    WWCC[L[j1],L[j2],L[j3],L[j4]]*
                    WWBB[L[k1],L[k2],L[k3],L[k4]],
  NewRule->True,Quiet->True],

ThreePerm=={2,1,3},
DefineRule[PermRule,Weyl[L[i1],L[i2],L[i3],L[i4]]*
                    Weyl[L[j1],L[j2],L[j3],L[j4]]*
                    Weyl[L[k1],L[k2],L[k3],L[k4]],
                    WWBB[L[i1],L[i2],L[i3],L[i4]]*
                    WWAA[L[j1],L[j2],L[j3],L[j4]]*
                    WWCC[L[k1],L[k2],L[k3],L[k4]],
  NewRule->True,Quiet->True],

ThreePerm=={2,3,1},
DefineRule[PermRule,Weyl[L[i1],L[i2],L[i3],L[i4]]*
                    Weyl[L[j1],L[j2],L[j3],L[j4]]*
                    Weyl[L[k1],L[k2],L[k3],L[k4]],
                    WWCC[L[i1],L[i2],L[i3],L[i4]]*
                    WWAA[L[j1],L[j2],L[j3],L[j4]]*
                    WWBB[L[k1],L[k2],L[k3],L[k4]],
  NewRule->True,Quiet->True],

ThreePerm=={3,1,2},
DefineRule[PermRule,Weyl[L[i1],L[i2],L[i3],L[i4]]*
                    Weyl[L[j1],L[j2],L[j3],L[j4]]*
                    Weyl[L[k1],L[k2],L[k3],L[k4]],
                    WWBB[L[i1],L[i2],L[i3],L[i4]]*
                    WWCC[L[j1],L[j2],L[j3],L[j4]]*
                    WWAA[L[k1],L[k2],L[k3],L[k4]],
  NewRule->True,Quiet->True],

ThreePerm=={3,2,1},
DefineRule[PermRule,Weyl[L[i1],L[i2],L[i3],L[i4]]*
                    Weyl[L[j1],L[j2],L[j3],L[j4]]*
                    Weyl[L[k1],L[k2],L[k3],L[k4]],
                    WWCC[L[i1],L[i2],L[i3],L[i4]]*
                    WWBB[L[j1],L[j2],L[j3],L[j4]]*
                    WWAA[L[k1],L[k2],L[k3],L[k4]],
  NewRule->True,Quiet->True]

]; (* End Which *)

Temp=OldTensor;
Temp=(Temp/.PermRule);
Temp=SuperSimplify[Temp];
Temp=TensorSimplify[Temp];
If[(Head[FormFlag]==Symbol)&&
   (ToString[FormFlag]=="False"),
  Null,
  Temp=(Temp/.ReturnThreeWeyl)];
Temp=NewDummy[Temp];
NewTensor=Temp;
ReturnValue=(TenExp+NewTensor-OldTensor//TensorExpand);
ReturnValue
)
]; (* End Module *)

(************ End PermWeyl ************)


(***********************************************************)
(*                        PermPpDD                         *)
(***********************************************************)
(* Function to permute copies of the tensor P_{ij|k} in a given
   term.  This function assumes that there are three copies
   of this tensor in the term or terms of interest.  If this
   is not the case, the function will not do any damage.

Originated on 2 June 2004

I tested this function carefully on 2 June 2004.

Required parameters:

(1) A tensor expression
(2) A permutation of the numbers 1, 2, and 3 (in the form of a
    list)

Optional parameters:

(3) A list or a single integer indicating which term or terms of the
    tensor expression the function is to act on.  Default value is
    "Omitted", which indicates that the function is to act on all
    terms of the expression.  The caller should code this as
    TermList->{2,5,9} (for example) or TermList->3 (for example).
(4) RetPpDD flag.  If set to "True", then, upon completion of
    the main work, the function will replace AAAAPD, BBBBPD, and
    CCCCPD with the Weyl tensor.  Default value is "True".

Return value:

A tensor expression which is equivalent to parameter (1).

*)

DefineTensor[AAAAPD,2,Symmetries->Symmetric];
DefineTensor[BBBBPD,2,Symmetries->Symmetric];
DefineTensor[CCCCPD,2,Symmetries->Symmetric];
DefineRule[ReturnThreePpDD,AAAAPD,P,NewRule->True];
DefineRule[ReturnThreePpDD,BBBBPD,P];
DefineRule[ReturnThreePpDD,CCCCPD,P];

(**** Main function definition ****)

Options[PermPpDD] := {RetPpDD->True,TermList->"Omitted"};

PermPpDD[TenExp_,ThreePerm_,opts___]:=

Module[
{
FormFlag,
LocalTermList,
NewTensor,
OldTensor,
PermRule,
ReturnValue,
Temp,
ValResult
},

(

ReturnValue=TenExp;

FormFlag=(RetPpDD/.{opts}/.Options[PermPpDD]);
LocalTermList=(TermList/.{opts}/.Options[PermPpDD]);

ValResult=TracCalcValSelect[TenExp,TermList->LocalTermList];

If[ValResult[[1]]!=0,Return[ReturnValue],Null];

OldTensor=ValResult[[2]];

If[ToString[Head[ThreePerm]]!="List",
(Message[TrCalc::HHHH];Return[ReturnValue]),
Null];

If[Length[ThreePerm]!=3,
(Message[TrCalc::IIII];Return[ReturnValue]),
Null];

If[Sort[ThreePerm]!={1,2,3},
(Message[TrCalc::JJJJ];Return[ReturnValue]),
Null];

Which[

ThreePerm=={1,2,3},
DefineRule[PermRule,P[L[i1],L[i2]][L[i3]]*
                    P[L[j1],L[j2]][L[j3]]*
                    P[L[k1],L[k2]][L[k3]],
               AAAAPD[L[i1],L[i2]][L[i3]]*
               BBBBPD[L[j1],L[j2]][L[j3]]*
               CCCCPD[L[k1],L[k2]][L[k3]],
  NewRule->True,Quiet->True],

ThreePerm=={1,3,2},
DefineRule[PermRule,P[L[i1],L[i2]][L[i3]]*
                    P[L[j1],L[j2]][L[j3]]*
                    P[L[k1],L[k2]][L[k3]],
               AAAAPD[L[i1],L[i2]][L[i3]]*
               CCCCPD[L[j1],L[j2]][L[j3]]*
               BBBBPD[L[k1],L[k2]][L[k3]],
  NewRule->True,Quiet->True],

ThreePerm=={2,1,3},
DefineRule[PermRule,P[L[i1],L[i2]][L[i3]]*
                    P[L[j1],L[j2]][L[j3]]*
                    P[L[k1],L[k2]][L[k3]],
               BBBBPD[L[i1],L[i2]][L[i3]]*
               AAAAPD[L[j1],L[j2]][L[j3]]*
               CCCCPD[L[k1],L[k2]][L[k3]],
  NewRule->True,Quiet->True],

ThreePerm=={2,3,1},
DefineRule[PermRule,P[L[i1],L[i2]][L[i3]]*
                    P[L[j1],L[j2]][L[j3]]*
                    P[L[k1],L[k2]][L[k3]],
               CCCCPD[L[i1],L[i2]][L[i3]]*
               AAAAPD[L[j1],L[j2]][L[j3]]*
               BBBBPD[L[k1],L[k2]][L[k3]],
  NewRule->True,Quiet->True],

ThreePerm=={3,1,2},
DefineRule[PermRule,P[L[i1],L[i2]][L[i3]]*
                    P[L[j1],L[j2]][L[j3]]*
                    P[L[k1],L[k2]][L[k3]],
               BBBBPD[L[i1],L[i2]][L[i3]]*
               CCCCPD[L[j1],L[j2]][L[j3]]*
               AAAAPD[L[k1],L[k2]][L[k3]],
  NewRule->True,Quiet->True],

ThreePerm=={3,2,1},
DefineRule[PermRule,P[L[i1],L[i2]][L[i3]]*
                    P[L[j1],L[j2]][L[j3]]*
                    P[L[k1],L[k2]][L[k3]],
               CCCCPD[L[i1],L[i2]][L[i3]]*
               BBBBPD[L[j1],L[j2]][L[j3]]*
               AAAAPD[L[k1],L[k2]][L[k3]],
  NewRule->True,Quiet->True]

]; (* End Which *)

Temp=OldTensor;
Temp=(Temp/.PermRule);
Temp=SuperSimplify[Temp];
Temp=TensorSimplify[Temp];
If[(Head[FormFlag]==Symbol)&&
   (ToString[FormFlag]=="False"),
  Null,
  Temp=(Temp/.ReturnThreePpDD)];
Temp=NewDummy[Temp];
NewTensor=Temp;
ReturnValue=(TenExp+NewTensor-OldTensor//TensorExpand);
ReturnValue
)
]; (* End Module *)

(************ End PermPpDD ************)


(***********************************************************)
(*                        ToWeylRho                        *)
(***********************************************************)
(*
Date last modified:  9 November 2003
  I tested this function on 9 November 2003.

Function receives a tensor expression and converts any
occurrences of Rm, Rc, or Sc to Weyl and Rho tensors.
*)

ToWeylRho[TenExp_]:=
Module[
{Temp},
(
 Temp=TenExp;
 Temp=(Temp/.RmToWeyl);
 Temp=(Temp/.RcToP);
 Temp=(Temp/.ScToTrP);
 Temp=(Temp/.TrPToInd);
 Temp
)];
(********** End of ToWeylRho *********)

(***********************************************************)
(*                        ToRmRcSc                         *)
(***********************************************************)
(*
Date last modified:  9 November 2003
  I tested this function on 9 November 2003.

Function receives a tensor expression and converts any
occurrences of Weyl, Rho, or TrP to Rm, Rc, and Sc.
*)

ToRmRcSc[TenExp_]:=
Module[
{Temp},
(
 Temp=TenExp;
 Temp=(Temp/.WeylToRm);
 Temp=(Temp/.TrPToInd);
 Temp=(Temp/.PToRc);
 Temp
)];
(********* End of ToRmRcSc *********)

(***********************************************************)
(*              RepeatFirstBianchiRule                     *)
(***********************************************************)

(* RepeatFirstBianchiRule

Date last modified:  17 December 2003

Function receives a tensor expression and applies the first Bianchi
rule to it.  It does this n times, where n is a number supplied by the
caller

*)

RepeatFirstBianchiRule[TensorExpression_,n_]:=

Module[

{
i,
Temp
},

(

If[ToString[Head[n]]!="Integer",
  (
   Message[TrCalc::KKKK];
   Return[TensorExpression]
  ),
  Null
  ];  (* End If *)

If[n<1,
  (
   Message[TrCalc::LLLL];
   Return[TensorExpression]
  ),
  Null
  ];  (* End If *)

Temp=TensorExpression;
Do[Temp=(Temp/.FirstBianchiRule),{i,1,n}];
Temp

)     (* End Module *)
];

(******** End RepeatFirstBianchiRule *************)

(***************************************************************)
(*                         SwapWeyl                            *)
(***************************************************************)
(* SwapWeyl:  Swap (interchange) copies of Weyl curvature tensor.

Date last modified:  21 December 2003
   I tested this function carefully on 22 December 2003.

Input:

Operand:  A tensor expression (which may include tractors)

TermList:  Parameter is optional.  TermList is a positive integer or a
  list of such integers indicating the terms to which the user wishes to
  apply the function.  If omitted, or if the user codes "Omitted", then
  SwapWeyl acts on all of Operand.

Urgency:  Parameter is optional.  Default value is 2, which means that
  SwapWeyl will apply the SuperSimplify command to its intermediate
  result.  The other option available to the caller is to specify
  Urgency=1.  In this case SwapWeyl will not call SuperSimplify.  This
  speeds up the computation but reduces the likelihood that SwapWeyl
  will successfully interchange the copies of the Weyl curvature
  tensor.  See remarks below.

Output:

A tensor expression equivalent to Operand.

At times one may want to change the order in which Ricci lists the
factors of some term of a tensor expression.  This problem occurs when
the same tensor appears more than once as a factor of a given term.
This function attempts to interchange two copies of the
undifferentiated Weyl curvature tensor which appear as factors.
If a term has more than two such factors, then the outcome of this
function is unpredictable.  In any case, this function will return a
tensor-tractor expression which is equivalent to Operand.  The user,
of course, must specify legitimate values for the parameters.

Ricci has certain procedures for deciding upon the order in which it
arranges factors of tensor expressions.  These rules may prevent this
function from successfully arranging the factors in the order
requested by the caller.  In any case, this function always returns a
tensor expression which is equivalent to Operand.

*)

DefineTensor[AAWeyl,4,Symmetries->RiemannSymmetries];
DefineRule[WeylWeylToWeylAA,
    Weyl[L[i1],L[i2],L[i3],L[i4]]Weyl[L[j1],L[j2],L[j3],L[j4]],
  AAWeyl[L[j1],L[j2],L[j3],L[j4]]Weyl[L[i1],L[i2],L[i3],L[i4]],
  NewRule->True];
DefineRule[AAToWeyl,AAWeyl,Weyl,NewRule->True];

Options[SwapWeyl] := {TermList->"Omitted",Urgency->2};

SwapWeyl[Operand_,opts___] :=

Module[{
  ValResult,
  Temp,
  OldTensor,
  NewTensor,
  ResultTensor,
  LocalUrgency,
  LocalTermList
},
(
  LocalUrgency=(Urgency/.{opts}/.Options[SwapWeyl]);
  LocalTermList=(TermList/.{opts}/.Options[SwapWeyl]);
  
  If[(ToString[Head[LocalUrgency]] != "Integer")||
     (LocalUrgency<1)||
     (LocalUrgency>2),
    (
    Message[TrCalc::FFFF];
    Return[Operand]
    ),
    Null
  ];

  ValResult=TracCalcValSelect[Operand,TermList->LocalTermList];

  If[ValResult[[1]]!=0,
    Return[Operand],
    (
    OldTensor=ValResult[[2]];
    Temp=OldTensor;
    Temp=(Temp/.WeylWeylToWeylAA);
    Temp=TensorSimplify[Temp];
    If[LocalUrgency==2,
      Temp=SuperSimplify[Temp],
      Null];
    Temp=(Temp/.AAToWeyl);
    Temp=NewDummy[Temp];
    NewTensor=Temp;
    ResultTensor=(Operand+NewTensor-OldTensor//TensorExpand);
    Return[ResultTensor]
    )
  ]   (* End If *)
)     (* End body of Module *)
];    (* End Module *)


(***************************************************************)
(*                         SwapPpDD                            *)
(***************************************************************)
(* SwapPpDD:  Swap (interchange) copies of P_{ij|k}.

24 June 2004 Changing Options[SwapWeyl] to Options[SwapPpDD].

I tested this function carefully on 2 June 2004.

Date last modified:  2 June 2004

Input:

Operand:  A tensor expression (which may include tractors)

TermList:  Parameter is optional.  TermList is a positive integer or a
  list of such integers indicating the terms to which the user wishes to
  apply the function.  If omitted, or if the user codes "Omitted", then
  SwapPpDD acts on all of Operand.

Urgency:  Parameter is optional.  Default value is 2, which means that
  SwapPpDD will apply the SuperSimplify command to its intermediate
  result.  The other option available to the caller is to specify
  Urgency=1.  In this case SwapPpDD will not call SuperSimplify.  This
  speeds up the computation but reduces the likelihood that SwapPpDD
  will successfully interchange the copies of the tensor.  See remarks
  below.

Output:

A tensor expression equivalent to Operand.

Ricci has certain procedures for deciding upon the order in which it
arranges factors of tensor expressions.  These rules may prevent this
function from successfully arranging the factors in the order
requested by the caller.  In any case, this function always returns a
tensor expression which is equivalent to Operand.

*)

DefineTensor[AAPpDD,2,Symmetries->Symmetric];
DefineRule[PpDDPpDDToPpDDAAPpDD,
  P[L[i1],L[i2]][L[i3]]     P[L[j1],L[j2]][L[j3]],
  P[L[i1],L[i2]][L[i3]]AAPpDD[L[j1],L[j2]][L[j3]],
  NewRule->True];
DefineRule[AAPpDDToPpDD,AAPpDD,P,NewRule->True];

Options[SwapPpDD] := {TermList->"Omitted",Urgency->2};

SwapPpDD[Operand_,opts___] :=

Module[{
  ValResult,
  Temp,
  OldTensor,
  NewTensor,
  ResultTensor,
  LocalUrgency,
  LocalTermList
},
(
  LocalUrgency=(Urgency/.{opts}/.Options[SwapPpDD]);
  LocalTermList=(TermList/.{opts}/.Options[SwapPpDD]);
  
  If[(ToString[Head[LocalUrgency]] != "Integer")||
     (LocalUrgency<1)||
     (LocalUrgency>2),
    (
    Message[TrCalc::FFFF];
    Return[Operand]
    ),
    Null
  ];

  ValResult=TracCalcValSelect[Operand,TermList->LocalTermList];

  If[ValResult[[1]]!=0,
    Return[Operand],
    (
    OldTensor=ValResult[[2]];
    Temp=OldTensor;
    Temp=(Temp/.PpDDPpDDToPpDDAAPpDD);
    Temp=TensorSimplify[Temp];
    If[LocalUrgency==2,
      Temp=SuperSimplify[Temp],
      Null];
    Temp=(Temp/.AAPpDDToPpDD);
    Temp=NewDummy[Temp];
    NewTensor=Temp;
    ResultTensor=(Operand+NewTensor-OldTensor//TensorExpand);
    Return[ResultTensor]
    )
  ]   (* End If *)
)     (* End body of Module *)
];    (* End Module *)

(***************************************************************)
(*                  WeylSecondBianchiRule                      *)
(***************************************************************)
(*

Date last modified:  7 January 2004

I tested this function on 8 January 2004.

Takes a derivative of the Weyl tensor, converts it to the Riemannian
curvature tensor (adding appropriate extra terms), and moves the last
three indices (two to the left of the semicolon and one to the right
of it) according to the second Bianchi rule.

Input:

Operand:  A tensor expression (which may include tractors)

TermList:  Parameter is optional.  TermList is a positive integer or a
  list of such integers indicating the terms to which the user wishes to
  apply the function.  If omitted, or if the user codes "Omitted", then
  function acts on all of Operand.

Output:

A tensor expression equivalent to Operand.

*)

(******************************************************)
DefineRule[ExpandWeylSecondBianchiRule,
Weyl[L[i],L[j],L[k],L[l]][L[m]],
   g[L[j], L[m]]*P[L[i], L[k]] [L[l]] - 
   g[L[j], L[l]]*P[L[i], L[k]] [L[m]] - g[L[j], L[m]]*P[L[i], L[l]] [L[k]] + 
   g[L[j], L[k]]*P[L[i], L[l]] [L[m]] + g[L[j], L[l]]*P[L[i], L[m]] [L[k]] - 
   g[L[j], L[k]]*P[L[i], L[m]] [L[l]] - g[L[i], L[m]]*P[L[j], L[k]] [L[l]] + 
   g[L[i], L[l]]*P[L[j], L[k]] [L[m]] + g[L[i], L[m]]*P[L[j], L[l]] [L[k]] - 
   g[L[i], L[k]]*P[L[j], L[l]] [L[m]] - g[L[i], L[l]]*P[L[j], L[m]] [L[k]] + 
   g[L[i], L[k]]*P[L[j], L[m]] [L[l]] + Weyl[L[i], L[j], L[k], L[m]] [L[l]] - 
   Weyl[L[i], L[j], L[l], L[m]] [L[k]],
NewRule->True,Quiet->True];
(******************************************************)

Options[WeylSecondBianchiRule] := {TermList->"Omitted"};

WeylSecondBianchiRule[Operand_,opts___] :=

Module[{
LocalTermList,
NewTensor,
OldTensor,
ResultTensor,
Temp,
ValResult
},
(
  LocalTermList=(TermList/.{opts}/.Options[WeylSecondBianchiRule]);
  ValResult=TracCalcValSelect[Operand,TermList->LocalTermList];

  If[ValResult[[1]]!=0,
    Return[Operand],
    (
    OldTensor=ValResult[[2]];
    Temp=OldTensor;
    Temp=(Temp/.ExpandWeylSecondBianchiRule);
    Temp=TensorSimplify[Temp];
    NewTensor=Temp;
    ResultTensor=(Operand+NewTensor-OldTensor//TensorExpand);
    Return[ResultTensor]
    )
  ]   (* End If *)
)     (* End body of Module *)
];    (* End Module *)


(******************************************************)
(*               SelectBianchiRules                   *)
(******************************************************)
(*
Date last modified:  26 May 2004

26 May 2004:  Copying over from earlier file.  Did a few tests of
              this function, but it might be a good idea to do more
              tests.

Input:

Operand:  a tensor or tractor expression
IndOnd,
IndTwo,
IndThree:  the three indices to be specified in the BianchiRules
  call
TermList (optional):  a positive integer or a list of such integers
  indicating the terms of Operand on which the function is to act.

In standard Ricci, the BianchiRules command is an ordinary rule which
one can apply to a tensor expression.  As of July 2002, standard Ricci
does not give the user the ability to apply BianchiRules to a proper
subset of the set of terms of an expression.  The present function
fills this gap.  It allows the user to apply BianchiRules to only
selected terms of the tensor expression.

*)

Options[SelectBianchiRules] := {TermList->"Omitted"};

SelectBianchiRules[Operand_,IndOne_,IndTwo_,IndThree_,opts___] :=

Module[
{
ValResult,
OldTensor,
NewTensor,
LocalTermList,
Temp,
ResultTensor
},
(
LocalTermList=(TermList/.{opts}/.Options[SelectBianchiRules]);

ValResult=TracCalcValSelect[Operand,TermList->LocalTermList];

If[ValResult[[1]]!=0,
  Return[Operand],
  Null
];

OldTensor=ValResult[[2]];
Temp=OldTensor;
Temp=(Temp/.BianchiRules[IndOne,IndTwo,IndThree]);
NewTensor=Temp;
ResultTensor=(Operand+NewTensor-OldTensor//TensorExpand);
Return[ResultTensor]
)
];

(******************************************************)

(******************************************************)
(*                   SelectRule                       *)
(******************************************************)
(*
Date last modified:  5 June 2004

Input:

Operand:  a tensor or tractor expression
RuleParam:  a rule that the caller has defined by DefineRule
TermList (optional):  a positive integer or a list of such integers
  indicating the terms of Operand on which the function is to act.

Output:

A tensor/tractor expression as modified by the application of the rule
to the indicated terms

The main point here is that the caller supplies a tensor/tractor
expression together with a rule and a list of terms.  SelectRule
applies the rule to the selected terms.

*)

Options[SelectRule] := {TermList->"Omitted"};

SelectRule[Operand_,RuleParam_,opts___] :=

Module[
{
ValResult,
OldTensor,
NewTensor,
LocalTermList,
Temp,
ResultTensor
},
(
LocalTermList=(TermList/.{opts}/.Options[SelectRule]);

ValResult=TracCalcValSelect[Operand,TermList->LocalTermList];

If[ValResult[[1]]!=0,
  Return[Operand],
  Null
];

OldTensor=ValResult[[2]];
Temp=OldTensor;
Temp=(Temp/.RuleParam);
NewTensor=Temp;
ResultTensor=(Operand+NewTensor-OldTensor//TensorExpand);
Return[ResultTensor]
)
];

(******************************************************)

(***   End of file ***)
